
//
// Copyright (C) 2018 David Carpenter
// 
// Permission is hereby granted, free of charge, 
// to any person obtaining a copy of this software 
// and associated documentation files (the "Software"), 
// to deal in the Software without restriction, 
// including without limitation the rights to use, 
// copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, 
// subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be 
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//


// compiler options
#pragma warning(disable : 4311) // type cast void * to unsigned int
#pragma warning(disable : 4312) // type cast unsigned int to void *
#pragma warning(disable : 4244) // warning C4244: 'argument' : conversion from 'LONG_PTR' to 'LONG', possible loss of data
#pragma warning(disable : 4996) // deprecation


#define _WIN32_IE 0x501
#define _WIN32_WINNT 0x501

#include <windows.h>
#include <stdio.h>

static int voidsync_get_registry_string(HKEY hkey,const wchar_t *value,wchar_t *wbuf,int size_in_wchars);
static int voidsync_set_registry_string(HKEY hkey,const wchar_t *value,const wchar_t *wbuf);

static SYSTEMTIME localst;
static FILE *logf = 0;

static int voidsync_get_registry_dword(HKEY hkey,const wchar_t *value,DWORD *dw)
{
	DWORD cbData;
	DWORD type;
	DWORD dwbuf;
	
	cbData = sizeof(DWORD);
	
	if (RegQueryValueExW(hkey,value,0,&type,(BYTE *)&dwbuf,&cbData) == ERROR_SUCCESS)
	{
		if (type == REG_DWORD)
		{
			*dw = dwbuf;
			
			return 1;
		}
	}
	
	return 0;
}

static int voidsync_set_registry_dword(HKEY hkey,const wchar_t *value,DWORD dw)
{
	if (RegSetValueExW(hkey,value,0,REG_DWORD,(BYTE *)&dw,sizeof(DWORD)) == ERROR_SUCCESS)
	{
		return 1;
	}
	
	return 0;
}

/*
typedef struct filename_s
{
};
*/

static int cmpft(FILETIME *ft1,FILETIME *ft2)
{
	if (ft1->dwHighDateTime < ft2->dwHighDateTime) return -1;
	if (ft1->dwHighDateTime > ft2->dwHighDateTime) return 1;

	if (ft1->dwLowDateTime < ft2->dwLowDateTime) return -1;
	if (ft1->dwLowDateTime > ft2->dwLowDateTime) return 1;
	
	return 0;
}		

void makepathtofile(wchar_t *file)
{
	wchar_t *p;
	wchar_t buf[MAX_PATH * 3];
	wchar_t *d;
		
	p = file;
	d = buf;

	// skip starting \\ in \\server\share	
	while(*p)
	{
		if (*p != '\\')
		{
			break;
		}
		
		*d++ = *p;
		p++;
	}	
	
	while(*p)
	{
		if (*p == '\\')
		{
			*d = 0;
			CreateDirectory(buf,0);
		}
		
		*d++ = *p;
		p++;
	}
}

void version_control_backup(wchar_t *dst,wchar_t *vc)
{
//	wchar_t vcbuf[MAX_PATH*3];
//	printf("%S => %S\n",dst,vc);

	makepathtofile(vc);
	if (CopyFile(dst,vc,FALSE))
	{
//		SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE),FOREGROUND_GREEN|FOREGROUND_INTENSITY);
//		printf("backup successful %S to %S\n",dst,vc);
//		if (logf) fprintf(logf,"<font color=\"#008000\">%d:%02d %d/%02d/%d backup successful %S to %S</font></br>\r\n",localst.wHour,localst.wMinute,localst.wDay,localst.wMonth,localst.wYear,dst,vc);
	}
	else
	{
		SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE),FOREGROUND_RED|FOREGROUND_INTENSITY);
		printf("%d: failed to backup: %S\n",GetLastError(),dst);

		if (logf) fprintf(logf,"<font color=\"#800000\">%d:%02d %d/%02d/%d %d: failed to backup: %S</font></br>\r\n",localst.wHour,localst.wMinute,localst.wDay,localst.wMonth,localst.wYear,GetLastError(),dst);
	}
	
//	wsprintf(dstbuf,L"%s\\%s",dst,srcfd.cFileName);
//	vcbuf
	
	//CreateDirectory.
}

int scan(wchar_t *src,wchar_t *dst,wchar_t *vc)
{	
	HANDLE h;
	WIN32_FIND_DATA srcfd;
	WIN32_FIND_DATA dstfd;
	wchar_t srcbuf[MAX_PATH*3];
	wchar_t dstbuf[MAX_PATH*3];
	wchar_t vcbuf[MAX_PATH*3];
	int versioncontrol;

//	SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE),FOREGROUND_BLUE|FOREGROUND_INTENSITY);
//	printf("scan %S %S\n",src,dst);

	versioncontrol = 1;

	wsprintf(srcbuf,L"%s\\_voidsync_noversioncontrol_",src);
	if (GetFileAttributes(srcbuf) != INVALID_FILE_ATTRIBUTES)
	{
		versioncontrol = 0;
	}

	wsprintf(srcbuf,L"%s\\*.*",src);
	h = FindFirstFile(srcbuf,&srcfd);
	if (h != INVALID_HANDLE_VALUE)
	{
		for(;;)
		{
			if (srcfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				if (wcscmp(srcfd.cFileName,L".") != 0)
				{
					if (wcscmp(srcfd.cFileName,L"..") != 0)
					{
						if (wcscmp(srcfd.cFileName,L"_VOIDSYNC_") != 0)
						{
							wsprintf(srcbuf,L"%s\\%s",src,srcfd.cFileName);
							wsprintf(dstbuf,L"%s\\%s",dst,srcfd.cFileName);
							wsprintf(vcbuf,L"%s\\%s",vc,srcfd.cFileName);
							
							makepathtofile(dstbuf);
							CreateDirectory(dstbuf,0);
							
							scan(srcbuf,dstbuf,vcbuf);
						}
					}
				}
			}
			else
			{
				int dstexists;
				SYSTEMTIME dstst;
				SYSTEMTIME dstlocalst;
				
				wsprintf(srcbuf,L"%s\\%s",src,srcfd.cFileName);
				wsprintf(dstbuf,L"%s\\%s",dst,srcfd.cFileName);

				dstexists = 1;
				
				if (!GetFileAttributesEx(dstbuf,GetFileExInfoStandard,&dstfd))
				{
					dstfd.ftLastWriteTime.dwHighDateTime = 0;
					dstfd.ftLastWriteTime.dwLowDateTime = 0;
					dstexists = 0;
				}

				{
					FileTimeToSystemTime(&dstfd.ftLastWriteTime,&dstst);
					SystemTimeToTzSpecificLocalTime(0,&dstst,&dstlocalst);
				}
				
				wsprintf(vcbuf,L"%s\\%s\\%d%02d%02dT%02d%02d%02d\\%s",vc,srcfd.cFileName,dstlocalst.wYear,dstlocalst.wMonth,dstlocalst.wDay,dstlocalst.wHour,dstlocalst.wMinute,dstlocalst.wSecond,srcfd.cFileName);
				
				if ((srcfd.ftLastWriteTime.dwHighDateTime != dstfd.ftLastWriteTime.dwHighDateTime) || (srcfd.ftLastWriteTime.dwLowDateTime != dstfd.ftLastWriteTime.dwLowDateTime))
				{
//						printf("%S\n",srcfd.cFileName);
					
					// copy the dst file to the version control folder
					if (versioncontrol)
					{
						if (dstexists)
						{
							version_control_backup(dstbuf,vcbuf);
						}
					}
					
//	printf("%S => %S\n",srcbuf,dstbuf);

					// copy src to dst.
					makepathtofile(dstbuf);
					if (CopyFile(srcbuf,dstbuf,FALSE))
					{
						SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE),FOREGROUND_GREEN|FOREGROUND_INTENSITY);
						printf("synced: %S\n",srcbuf);
						
						if (logf) fprintf(logf,"<font color=\"#008000\">%d:%02d %d/%02d/%d synced: %S</font></br>\r\n",localst.wHour,localst.wMinute,localst.wDay,localst.wMonth,localst.wYear,srcbuf);
					}
					else
					{
						SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE),FOREGROUND_RED|FOREGROUND_INTENSITY);
						printf("%d: failed to copy %S\n",GetLastError(),srcbuf);

						if (logf) fprintf(logf,"<font color=\"#800000\">%d:%02d %d/%02d/%d %d: failed to copy: %S</font></br>\r\n",localst.wHour,localst.wMinute,localst.wDay,localst.wMonth,localst.wYear,GetLastError(),srcbuf);
					}						
				}

			/*
				char dst[MAX_PATH*3];
				
				sprintf(dst,"%s\\*.*",argv[1]);
				sprintf(filename,"D:\\Uploads\\MMS\\%S",fd.cFileName);
				
				f = CreateFile(filename,GENERIC_READ,FILE_SHARE_DELETE|FILE_SHARE_WRITE|FILE_SHARE_READ,0,OPEN_EXISTING,0,0);
				if (f != INVALID_HANDLE_VALUE)
				{
					size = GetFileSize(f,0);
					
					ReadFile(f,buf,size,&numread,0);
					
					p = buf;
					e = buf + size;
					while(p < e)
					{
						if ((*p == 'J') && (p[1] == 'F') && (p[2] == 'I') && (p[3] == 'F'))
						{
							DWORD numwritten;
							
							printf("found jfif at %d\n",p-buf);
							
							WriteFile(o,p-6,size-(p-buf)+6,&numwritten,0);
							
							break;
						}
						p++;
					}
					
					CloseHandle(o);
					CloseHandle(f);
				}
				else
				{
					printf("cant open\n");
				}*/
			}

			if (!FindNextFile(h,&srcfd)) break;
		}
		
		FindClose(h);
	}

	return 0;
}

void remove_trailing_backslash(wchar_t *s)
{
	wchar_t *p;
	
	p = s;
	
	while(*p)
	{
		if ((*p == '\\') && (p[1] == 0))
		{
			*p = 0;
			break;
		}
		
		p++;
	}
}

int main(int argc, char **argv)
{
	wchar_t srcbuf[MAX_PATH*3];
	wchar_t dstbuf[MAX_PATH*3];
	wchar_t vcbuf[MAX_PATH*3];
	FILETIME ft;

	argc--;
	if (argc != 2)
	{
		MessageBox(0,L"voidsync.exe <src> <dst>",L"voidsync",MB_OK);
		
		return 1;
	}
	
	GetSystemTimeAsFileTime(&ft);
	GetLocalTime(&localst);

	// log
	{
		char logfilename[MAX_PATH];
		
		wsprintf(dstbuf,L"%S",argv[2]);
		remove_trailing_backslash(dstbuf);
		sprintf(logfilename,"%S\\_VOIDSYNC_\\log.htm",dstbuf);
		
		logf = fopen(logfilename,"ab");
		
		if (logf) fprintf(logf,"<font color=\"#000080\">%d:%02d %d/%02d/%d sync started</font></br>\r\n",localst.wHour,localst.wMinute,localst.wDay,localst.wMonth,localst.wYear);
	}
	
/*	// load settings
	{
		char inifilename[MAX_PATH];
		
		wsprintf(dstbuf,L"%S",argv[2]);
		remove_trailing_backslash(dstbuf);
		sprintf(inifilename,"%S\\_VOIDSYNC_\\voidsync.ini",dstbuf);
		
		last_sync_time.dwHighDateTime = GetPrivateProfileIntA("voidsync","last_sync_time_high",0,inifilename);
		last_sync_time.dwLowDateTime = GetPrivateProfileIntA("voidsync","last_sync_time_low",0,inifilename);
	}
*/	
	/*
	{
		HKEY hkey;
		
		if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,L"Software\\voidSync",0,KEY_READ,&hkey) == ERROR_SUCCESS)
		{
			voidsync_get_registry_dword(hkey,L"last_sync_time_high",&last_sync_time.dwHighDateTime);
			voidsync_get_registry_dword(hkey,L"last_sync_time_low",&last_sync_time.dwLowDateTime);
			
			RegCloseKey(hkey);
		}
	}
	*/
	wsprintf(srcbuf,L"%S",argv[1]);
	wsprintf(dstbuf,L"%S",argv[2]);

	remove_trailing_backslash(srcbuf);
	remove_trailing_backslash(dstbuf);
		
	wsprintf(vcbuf,L"%s\\_VOIDSYNC_",dstbuf);
	
	scan(srcbuf,dstbuf,vcbuf);
	/*
	// save settings
	{
		char inifilename[MAX_PATH];
		char numbuf[256];
		
		wsprintf(dstbuf,L"%S",argv[2]);
		remove_trailing_backslash(dstbuf);
		sprintf(inifilename,"%S\\_VOIDSYNC_\\voidsync.ini",dstbuf);
		
		sprintf(numbuf,"%u",ft.dwHighDateTime);
		WritePrivateProfileStringA("voidsync","last_sync_time_high",numbuf,inifilename);
		sprintf(numbuf,"%u",ft.dwLowDateTime);
		WritePrivateProfileStringA("voidsync","last_sync_time_low",numbuf,inifilename);
	}
*/		
	// save settings
	/*
	{
		HKEY hkey;
		
		if (RegCreateKeyEx(HKEY_LOCAL_MACHINE,L"Software\\voidSync",0,NULL,REG_OPTION_NON_VOLATILE,KEY_WRITE,NULL,&hkey,0) == ERROR_SUCCESS)
		{
			voidsync_set_registry_dword(hkey,L"last_sync_time_high",ft.dwHighDateTime);
			voidsync_set_registry_dword(hkey,L"last_sync_time_low",ft.dwLowDateTime);

			RegCloseKey(hkey);
		}
	}*/
	
	if (logf)
	{
		fclose(logf);
	}

	return 0;
}
